""" This module implements a vehicle model.

    Vehicle model acts as an instance to modify vehicle's behaviour according to 
"""

from typing import Dict, List
from collections import OrderedDict
import itertools
import numpy as np
import pandas as pd

from symupy.utils import constants as ct

from .dynamics import VehicleDynamic


class Vehicle(object):
    """Class for defining a vehicle
    """

    counter = itertools.count()

    def __init__(
        self,
        abscisa=0.0,
        acceleration=0.0,
        distance=0.0,
        vehid=0,
        ordinate=0.0,
        link="",
        vehtype="",
        speed=0.0,
        lane=0,
        elevation=0.0,
        dynamic=VehicleDynamic(),
        itinerary=[],
    ):
        """ This initializer creates a Vehicle
        """
        self.abscisa = abscisa
        self.acceleration = acceleration
        self.distance = distance
        self.vehid = vehid
        self.ordinate = ordinate
        self.link = link
        self.vehtype = vehtype
        self.speed = speed
        self.lane = lane
        self.elevation = elevation
        self.dynamic = dynamic
        self.itinerary = itinerary

    def __repr__(self):
        data_dct = ", ".join(f"{k}:{v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({data_dct})"

    def __str__(self):
        data_dct = ", ".join(f"{k}:{v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({data_dct})"

    def update_state(self, dataveh):
        """Updates data within the structure with 
        
        :param dataveh: vehicle 
        :type dataveh: Vehicle
        """
        self.__dict__.update(**dataveh)

        link = getattr(self, "link")
        if link not in getattr(self, "itinerary"):
            self.itinerary.append(link)

    @property
    def C(self):
        """Output matrix"""
        return self.__output_matrix

    def activate_sensor(self, **kwargs) -> None:
        """Define the observation matrix and observed states in the vehicle
           
           vehicle.activate_sensor(speed=True,position=True)
        
        :return: Set the observation matrix __output_matrix
        :rtype: None
        """
        dct_idx = {"position": 0, "speed": 1, "acceleration": 2}
        C = np.zeros([len(kwargs.keys()), 3])
        for key in kwargs.keys():
            idx = dct_idx.get(key)
            C[idx][idx] = kwargs.get(key, 0)
        self.__output_matrix = C

    @property
    def vector_state(self) -> np.array:
        """Vehicle state vector (x,v,a)"""
        return np.array((self.distance, self.speed, self.acceleration))

    @property
    def observed_state(self) -> np.array:
        """ Return observed states via C@(x,v,a)"""
        return self.C @ self.vector_state

    def predict_state(self, control) -> np.array:
        """ Return predicted states via self.dynamic"""
        return self.dynamic(self, control)

    @staticmethod
    def format_dict(dataveh: OrderedDict) -> dict:
        """ This function creates the dictionary 
               {"abscisa":       float(data),
                "acceleration":  float(data),
                "distance":      float(data),
                "vehid":         int(data),
                "ordinate":      float(data),
                "link":          str (data),
                "vehtype":       str (data),
                "speed":         float(data),
                "lane":          int (data),
                "elevation":     float(data),
               } 
        
        :param dataveh: Ordered Dictionary from XML query
        :type dataveh: OrderedDict
        :return: Dictionary as in description
        :rtype: [type]
        """
        data = {ct.FIELD_DATA[key]: ct.FIELD_FORMAT[key](val) for key, val in dataveh.items()}
        return data

    @classmethod
    def from_response(cls, dataveh: OrderedDict):
        """Constructor for the class from a specific dictionary
        
        :param dataveh: Ordered dictionary from XML query
        :type dataveh: OrderedDict
        :return: Vehicle object
        :rtype: [type]
        """
        return cls(**Vehicle.format_dict(dataveh))


lstordct = List[OrderedDict]
lstvehs = List[Vehicle]


class VehicleList(object):
    """Class for defining a list of vehicles
    """

    def __init__(self, newvehs: lstvehs):
        self.vehicles = {}
        for veh in newvehs:
            self.vehicles[veh.vehid] = veh

    def update_list(self, current_vehs: lstordct):
        """ Appends a new list of vehicles

        :param vehlist: List of vehicles
        :type vehlist: lstordct
        """
        # Reformating data in vehicle dict
        current_vehs = [Vehicle.format_dict(veh) for veh in current_vehs]

        for veh in current_vehs:
            veh_id = veh.get("vehid")
            if veh_id in self.vehicles.keys():
                # Update existing vehicle
                self.vehicles.get(veh_id).update_state(veh)
            else:
                # Create a new vehicle and append
                self.vehicles[veh_id] = Vehicle(**veh)

    def _get_vehicles_attribute(self, attribute: str) -> np.array:
        """ Retrieve list of parameters 
        
        :param attribute: One of the vehicles attribute e.g. 'distance'
        :type attribute: str
        :return: vector of all parameters
        :rtype: np.array
        """
        constructor, ftype = ct.FIELD_FORMATAGG[attribute]
        if ftype:
            return constructor([getattr(veh, attribute) for veh in self], dtype=ftype)
        return [getattr(veh, attribute) for veh in self]  # Case str

    @property
    def acceleration(self):
        return self._get_vehicles_attribute("acceleration")

    @property
    def speed(self):
        return self._get_vehicles_attribute("speed")

    @property
    def distance(self):
        return self._get_vehicles_attribute("distance")

    def _to_pandas(self) -> pd.DataFrame:
        """ Transforms vehicle list into a pandas for rendering purposes 
        
        :return: Returns a table with pandas data.
        :rtype: pd.DataFrame
        """

        df_print = pd.DataFrame()
        for key, value in self.vehicles.items():
            df_print = df_print.append(pd.DataFrame(value.__dict__, index=(key,)))
        return df_print

    def __str__(self):
        if not self.vehicles:
            return "No vehicles have been registered"
        df_to_print = self._to_pandas()
        return str(df_to_print)

    def __repr__(self):
        if not self.vehicles:
            return "No vehicles have been registered"
        df_to_print = self._to_pandas()
        return repr(df_to_print)

    def __iter__(self):
        self.iterveh = iter(self.vehicles.values())
        return self

    def __next__(self):
        return next(self.iterveh)

    def __contains__(self, veh: Vehicle):
        return veh.vehid in self.get_vehid

    def __getitem__(self, key: int):
        return self.vehicles[key]

    @classmethod
    def from_request(cls, vehlistdct: lstordct):
        """Constructs a vehicle list from a list of ordered dictionaries (simulator repsonse)

        :param vehlist: list containing vehicle data in ordered dict format
        :type vehlist: list
        :return: vehicle list containing list of vehicle data classes
        :rtype: VehicleList
        """
        return cls([Vehicle.from_response(d) for d in vehlistdct])

    @property
    def get_vehid(self):
        return [v.vehid for v in self.vehicles]



import typing
from typing import TypeVar, Iterable, Tuple, Union

# Local imports
from .road_networks import RoadSideUnit
from symupy.components.vehicles import Vehicle


class V2XNetwork(object):
    def __init__(self) -> None:
        self._nodes_elem = []
        self._veh_elem = []

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __str__(self) -> str:
        tp = tuple(veh.vehid for veh in self._veh_elem)
        return f"Vehicle registered: {tp}\n Elements registered: {tp}"

    def register_vehicle(self, veh: Vehicle) -> None:
        self._veh_elem.append(veh)


class V2INetwork(V2XNetwork):
    def __init__(self) -> None:
        super().__init__()

    def register_element(self, element: RoadSideUnit) -> None:
        self._nodes_elem.append(element)


class V2VNetwork(V2XNetwork):
    def __init__(self) -> None:
        super().__init__()



""" 
**Connector Module**

This module details the implementation of a ``Simulator`` object in charge of handling the connection between the traffic simulator and this interface. The connection with the traffic simulator is handled by an object called ``Connector`` which establishes a messaging protocol with the traffic simulator. 

Example:
    To use the ``Simulator`` declare in a string the ``path`` to the simulator ::

        >>> path = "path/to/simulator.so"
        >>> simulator = Simulator(path) 

Other parameters can also be send to the simulator in order to provide other configurations:

Example: 
    To send make increase the *buffer size* to a specific size:

        >>> simulator = Simulator(path, bufferSize = 1000000)
    
    To increase change the flag that traces the flow:

        >>> simulator = Simulator(path, traceFlow = True)


"""

# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

import os
from itertools import repeat
from ctypes import cdll, create_string_buffer, c_int, byref, c_bool, c_double

import typing
from typing import Union

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

# Error Handling
from symupy.utils.exceptions import (
    SymupyLoadLibraryError,
    SymupyFileLoadError,
    SymupyVehicleCreationError,
    SymupyDriveVehicleError,
    SymupyWarning,
)

#
from symupy.api.scenario import Simulation
from symupy.utils import SimulatorRequest, Configurator

from symupy.utils import timer_func, printer_time
from symupy.utils import constants as ct

from symupy.components import V2INetwork, V2VNetwork
from symupy.components import Vehicle

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================

# V2X Connectivity
NetworkType = Union[V2INetwork, V2VNetwork]


class Simulator(object):
    """ This object describes is a connector manager for the interface between the traffic simulator and the 
    
        :param object: [description]
        :type object: [type]
        :raises SymupyLoadLibraryError: Error raised whenever the SymuVia library is not found
        :raises SymupyFileLoadError: Error raised whenever the provided path for an scenario cannot be loaded into the Simulator
        :raises SymupyVehicleCreationError: Error raised when a vehicle cannot be created
        :raises SymupyDriveVehicleError: Error rased when a vehicle state cannot be imposed
        :raises NotImplementedError: Not implemented functionality 
        :return: [description]
        :rtype: Simulator
    """

    def __init__(self, libraryPath: str = "", **kwargs) -> None:
        self.initialize_configurator(libraryPath=libraryPath, **kwargs)
        self._net = []

    def __repr__(self):
        return f"{self.__class__.__name__}({self.libraryPath})"

    def initialize_configurator(self, pathSimulator: str = "", **kwargs) -> None:
        """ 
           This method initialize a ``Configurator`` class that contains a small summary setup to launch a simulation
        """
        self._config = Configurator(libraryPath=pathSimulator, **kwargs)

    def load_symuvia(self) -> None:
        """ load SymuVia shared library """
        try:
            lib_symuvia = cdll.LoadLibrary(self.libraryPath)
        except OSError:
            raise SymupyLoadLibraryError("Library not found", self.libraryPath)
        self._library = lib_symuvia

    def load_network(self) -> None:
        """ load SymuVia Simulation File """
        if not hasattr(self, "_sim"):
            raise SymupyFileLoadError("File not provided", "")
        valid = self._library.SymLoadNetworkEx(self._sim.filename_encoded)
        if not valid:
            raise SymupyFileLoadError("Simulation could not be loaded", "")

    def init_simulation(self) -> None:
        """ Initializes conditions for a step by step simulation"""
        # Pointers
        self._s_response = create_string_buffer(ct.BUFFER_STRING)
        self._b_end = c_int()
        self._b_trace = c_bool(False)
        # self._b_force = c_int(1)
        self.state = SimulatorRequest()

    @timer_func
    def run_simulation(self, sim_object: Simulation = "") -> None:
        """ Run simulation in a single shot

        Args:
            sim_object (Simulation): Valid simulation scenario

        Returns:
            None: No returns provided, only internal updates
        """
        if sim_object:
            self.register_simulation(sim_object)

        self.load_symuvia()
        self._library.SymRunEx(self._sim.filename_encoded)

    def register_simulation(self, sim_object: Simulation) -> None:
        """Register simulation file within the simulator

        :param sim_object: Simulation scenario to register
        :type sim_object: Simulation
        :return: No value is returned
        :rtype: None
        """
        self._sim = sim_object

    def register_network(self, network: NetworkType) -> None:
        self._net.append(network)

    def request_answer(self):
        """Request simulator answer and maps the data locally
        """
        if self._config.stepLaunchMode == "lite":
            self._bContinue = self._library.SymRunNextStepLiteEx(self._b_trace, byref(self._b_end))
            return
        self._bContinue = self._library.SymRunNextStepEx(self._s_response, self._b_trace, byref(self._b_end))
        self.state.parse_data(self.s_response_dec)

    # @printer_time
    def run_step(self) -> int:
        """ Run simulation step by step

        :return: iteration step
        :rtype: int
        """
        try:
            self.request_answer()
            self._c_iter = next(self._n_iter)
            return self._c_iter
        except StopIteration:
            self._bContinue = False
            return -1

    def stop_step(self):
        """Stop current current step of running simulation
        """
        self._bContinue = False

    def create_vehicle(self, vehtype: str, origin: str, destination: str, lane: int = 1, simid: int = 0) -> int:
        """Creates a vehicle within the network

        :param vehtype: vehicle type according to simulation definitions
        :type vehtype: str
        :param origin: network endpoint nodeaccording to simulation
        :type origin: str
        :param destination: network endpoint nodeaccording to simulation
        :type destination: str
        :param lane: vehicle lane number, defaults to 1
        :type lane: int, optional
        :param simid: simulation id, defaults to 0
        :type simid: int, optional
        :raises SymupyVehicleCreationError: Exception handling for invalid vehicle types or invalid network points
        :return: Vehicle id of the vehicle created >0
        :rtype: int
        """
        endpoints = self._sim.get_network_endpoints()
        veh_data = self._sim.get_vehicletype_information()
        dbTime = self._sim.time_step
        vehid = tuple(v["id"] for v in veh_data)
        if vehtype not in vehid:
            raise SymupyVehicleCreationError("Unexisting Vehicle Class in File: ", self._sim.filename)

        if (origin not in endpoints) or (destination not in endpoints):
            raise SymupyVehicleCreationError("Unexisting Network Endpoint File: ", self._sim.filename)

        vehid = self._library.SymCreateVehicleEx(
            vehtype.encode("UTF8"), origin.encode("UTF8"), destination.encode("UTF8"), c_int(lane), c_double(dbTime),
        )
        return vehid

    def create_vehicle_with_route(
        self, vehtype: str, origin: str, destination: str, lane: int = 1, creation_time: float = 0, route: str = ""
    ) -> int:
        """ Creates a vehicle with a specific route
        
            :param vehtype: vehicle type according to simulation definitions
            :type vehtype: str
            :param origin: network endpoint node according to simulation
            :type origin: str
            :param destination: network endpoint node according to simulation
            :type destination: str
            :param lane: vehicle lane number, defaults to 1
            :type lane: int, optional
            :param creation_time: time instant of creation [0,Ts], defaults to 0
            :type creation_time: float, optional
            :param route: route followed by the vehicle, defaults to ""
            :type route: str, optional

            :return: Vehicle id of the vehicle created >0
            :rtype: int        
        """
        if origin == destination:
            return -1

        endpoints = self._sim.get_network_endpoints()
        veh_data = self._sim.get_vehicletype_information()
        dbTime = self.simulationstep
        vehid = tuple(v["id"] for v in veh_data)

        if vehtype not in vehid:
            raise SymupyVehicleCreationError("Unexisting Vehicle Class in File: ", self._sim.filename)

        if (origin not in endpoints) or (destination not in endpoints):
            raise SymupyVehicleCreationError("Unexisting Network Endpoint File: ", self._sim.filename)

        vehid = self._library.SymCreateVehicleWithRouteEx(
            origin.encode("UTF8"),
            destination.encode("UTF8"),
            vehtype.encode("UTF8"),
            c_int(lane),
            c_double(creation_time - dbTime),
            route.encode("UTF8"),
        )
        return vehid

    def drive_vehicle(self, vehid: int, new_pos: float, destination: str = None, lane: str = 1) -> None:
        """Drives a vehicle to a specific position

        :param vehid: vehicle id to drive 
        :type vehid: int
        :param new_pos: position to place the vehicle
        :type new_pos: float
        :param destination: link of destination, defaults to None
        :type destination: str, optional
        :param lane: lane fo destination, defaults to 1
        :type lane: str, optional
        :raises SymupyDriveVehicleError: Raises error when link does not exist
        :return: Value returned by SymDriveErr 1 ok, negative values are errors.
        :rtype: None
        """
        links = self._sim.get_network_links()

        if not destination:
            destination = self.state.query_vehicle_link(str(vehid))[0]

        if destination not in links:
            raise SymupyDriveVehicleError("Unexisting Network Endpoint File: ", self._sim.filename)

        # TODO: Validate that position do not overpass the max pos
        dr_state = self._library.SymDriveVehicleEx(
            c_int(vehid), destination.encode("UTF8"), c_int(lane), c_double(new_pos), 1
        )
        self.request_answer()
        return dr_state

    def drive_vehicle_with_control(self, vehcontrol, vehid: int, destination: str = None, lane: str = 1):
        # TODO: Basic prototyping
        vehcontrol.set_current_state(self.state)
        new_pos = vehcontrol.new_position
        return self.drive_vehicle(vehid, new_pos, destination, lane)

    def get_vehicle_context(self, vehid: str):
        ## TODO: Implement this
        raise NotImplementedError

    def log_vehicle_in_network(self, veh: Vehicle, network: NetworkType):
        # veh = Vehicle(vehid)
        ## TODO: Finish
        network.register_vehicle(veh)

    def log_vehid_in_network(self, vehid: str, network: NetworkType):
        ## TODO: Optional
        pass

    def init_total_travel_time(self):
        """ Counter initializer for total travel time
        """
        # TODO: Improvement → Better organizadtion
        self._library.SymGetTotalTravelTimeEx.restype = c_double

    def init_total_travel_distance(self):
        """ Counter initializer for total travel time
        """
        # TODO: Improvement → Better organizadtion
        self._library.SymGetTotalTravelDistanceEx.restype = c_double

    def get_total_travel_time(self, zone_id: str = None):
        """ Computes the total travel time of vehicles in a MFD region
        
        :param zone_id: MFD sensor id, defaults to None
        :type zone_id: str, optional
        :return: Associated total travel time
        """
        # TODO: Improvement → Better organizadtion
        if zone_id:
            return self._library.SymGetTotalTravelTimeEx(zone_id.encode("UTF8"))

        sensors = self.simulation.get_mfd_sensor_names()
        return tuple(self._library.SymGetTotalTravelTimeEx(sensor.encode("UTF8")) for sensor in sensors)

    def get_total_travel_distance(self, zone_id: str = None):
        """ Computes the total travel distance of vehicles in a MFD region
        
        :param zone_id: MFD sensor id, defaults to None
        :type zone_id: str, optional
        :return: Associated total travel distance
        """
        # TODO: Improvement → Better organizadtion
        if zone_id:
            return self._library.SymGetTotalTravelDistanceEx(zone_id.encode("UTF8"))

        sensors = self.simulation.get_mfd_sensor_names()
        return tuple(self._library.SymGetTotalTravelDistanceEx(sensor.encode("UTF8")) for sensor in sensors)

    def get_mfd_speed(self, zone_id: str = None):
        """ Computes the total speed of vehicles in a MFD region
        
        :param zone_id: MFD sensor id, defaults to None
        :type zone_id: str, optional
        :return: speed computed as ttt/ttd
        """
        # TODO: Improvement → Better organizadtion
        if zone_id:
            d = self.get_total_travel_distance(zone_id)
            t = self.get_total_travel_time(zone_id)
            spd = d / t if t != 0 else 10
            return spd

        itdsttm = zip(self.get_total_travel_distance(), self.get_total_travel_time())
        spd = []
        for d, t in itdsttm:
            if t != 0:
                spd.append(d / t)
            else:
                spd.append(10)  # minimum speed?
        return tuple(spd)

    def add_control_probability_zone_mfd(self, access_probability: dict, minimum_distance: dict) -> None:
        self.dctidzone = {}

        for tp_zn_pb, tp_zn_md in zip(access_probability.items(), minimum_distance.items()):
            sensor, accrate = tp_zn_pb
            _, min_dst = tp_zn_md
            links = self.simulation.get_links_in_mfd_sensor(sensor)
            links_str = " ".join(links)
            self.dctidzone[sensor] = self._library.SymAddControlZoneEx(
                -1, c_double(accrate), c_double(min_dst), f"'{links_str}'".encode("UTF8"),
            )
        return self.dctidzone

    def modify_control_probability_zone_mfd(self, access_probability: dict) -> None:

        for sensor, probablity in access_probability.items():
            self._library.SymModifyControlZoneEx(-1, self.dctidzone[sensor], c_double(probablity))
        return self.dctidzone

    def __enter__(self) -> None:
        """ Implementation as a context manager
            FIXME: Implement state machine ???
        """
        self.load_symuvia()
        self.load_network()
        self.init_simulation()
        self._n_iter = iter(self._sim.get_simulation_steps())
        self._c_iter = next(self._n_iter)
        self._bContinue = True
        # Extra
        self.init_total_travel_distance()
        self.init_total_travel_time()
        self.build_dynamic_param()
        return self

    def __exit__(self, type, value, traceback) -> bool:
        return False

    def build_dynamic_param(self):
        """Construct parameters for vehicle dynamics
        """
        self.__dct_par = {
            "time_step": self.simulation.time_step,
            "engine_tau": ct.ENGINE_CONSTANT,
        }

    @property
    def s_response_dec(self):
        """ Obtains instantaneous data from simulator

        :return: last query from simulator
        :rtype: str
        """
        return self._s_response.value.decode("UTF8")

    @property
    def do_next(self) -> bool:
        return self._bContinue

    @property
    def get_request(self) -> dict:
        return self.state.data_query

    @property
    def libraryPath(self) -> str:
        return self._config.libraryPath

    @property
    def simulation(self) -> Simulation:
        return self._sim

    @property
    def casename(self) -> str:
        return self.simulation.filename

    @property
    def simulationstep(self) -> str:
        return self._c_iter

    @property
    def time_step(self) -> float:
        return self.simulation.time_step

    @property
    def sampling_time(self) -> float:
        return self.simulation.time_step

    @classmethod
    def from_path(cls, filename_path, simuvia_path):
        """ Alternative constructor for the Simulator 

            Simulator.from_path(file,lib)
        """
        case = Simulation(filename_path)
        sim = cls(simuvia_path)
        sim.register_simulation(case)
        return sim



""" 
    This module contains a base class definition representing a single generic state
"""

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


class State(object):
    """
        This class defines a state object which provides basic functionalities for individual states within the state machine. 
    """

    def __init__(self):
        print("State:", str(self))

    def on_event(self, event):
        """
        Handle events that are delegated to this State.
        """
        pass

    def __repr__(self):
        """
        Leverages the __str__ method to describe the State.
        """
        return self.__str__()

    def __str__(self):
        """
        Returns the name of the State.
        """
        return self.__class__.__name__



from symupy.components.networks import RoadNetwork, RoadSideUnit
from symupy.components.networks import V2INetwork, V2VNetwork
from symupy.components.vehicles import Vehicle, VehicleList
from symupy.components.control import VehicleControl, VehicleGroupControl



from functools import wraps

# Decorators


def logger_func(orig_func):
    import logging

    logging.basicConfig(filename="{}.log".format(orig_func.__name__), level=logging.INFO)

    @wraps(orig_func)
    def wrapper(*args, **kwargs):
        logging.info("Ran with args: {}, and kwargs: {}".format(args, kwargs))
        return orig_func(*args, **kwargs)

    return wrapper


def timer_func(orig_func):
    import time

    @wraps(orig_func)
    def wrapper(*args, **kwargs):
        t1 = time.time()
        result = orig_func(*args, **kwargs)
        t2 = time.time() - t1
        print("{} ran in: {} sec".format(orig_func.__name__, t2))
        return result

    return wrapper


def printer_time(orig_func):
    @wraps(orig_func)
    def wrapper(*args, **kwargs):
        print(f"Step: {args[0].simulationstep}")
        return orig_func(*args, **kwargs)

    return wrapper



"""
    This module contains a set of exceptions and handlers for such exceptions.

    Example: 
        To use the ``Exceptions`` import the module as::

            >>> from symupy.utils.exceptions import SymupyLoadLibraryError
            >>> from ctypes import cdll
            >>> try: 
            ...     path = 'path/to/library'
            ...     lib = cdll.LoadLibrary(path)
            ... except OSError: 
            ...     raise  SymupyLoadLibraryError("Library not found",path)

"""
# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

import warnings

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


class SymupyError(Exception):
    """ 
        General error exception, it can be raised at whatever moment, the class its intends to create a base class that intends to tag error situations dected by this module. 
    
        :return: General ``SymupyError`` exception handler
        :rtype: SymupyError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message)
        self._target_dir = target_dir

    @property
    def target_dir(self) -> str:
        """ Retreives the path of the target directory
        
        :return: Path to target directory
        :rtype: str
        """
        return self._target_dir

    @property
    def get_messsage(self) -> str:
        """ Retreives custom message to be send to the user 
        
        :return: Custom message
        :rtype: str
        """
        (mess,) = self.args
        return mess


class SymupyWarning(object):
    """
       General warning exception, it can be raised at whatever moment, the class its intends to create a base class that intends to tag warning situations dected by this module. 
    
        :param Exception: General ``SymupyWarning`` exception handler
        :type Exception: SymupyWarning
    """

    def __init__(self, warning_message: str) -> None:
        warnings.warn(warning_message)


class SymupyFileLoadError(SymupyError):
    """
        File Load Error exception handler, created for handling file situations in particular scenario situations
    
        :return: File Load Error 
        :rtype: SymupyFileLoadError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)

    def __str__(self) -> str:
        return f"{self.get_messsage}, at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyFileLoadError({self.get_messsage},{self._target_dir})"


class SymupyLoadLibraryError(SymupyError):
    """
        Load Library Error exception handler, created for handling library situations in particular when it is not 
        possible to link the library and python
    
        :return: Library Load Error 
        :rtype: SymupyLoadLibraryError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)

    def __str__(self) -> str:
        return f"{self.get_messsage}, at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyLoadLibraryError({self.get_messsage},{self._target_dir})"


class SymupyVehicleCreationError(SymupyError):
    """
        Vehicle Creation Error exception handler, created for handling situations at runtime when the vehicles cannot be created within the network
    
        :return: Vehicle Creation Error 
        :rtype: SymupyVehicleCreationError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)

    def __str__(self) -> str:
        return f"{self.get_messsage}, at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyVehicleCreationError({self.get_messsage},{self._target_dir})"


class SymupyDriveVehicleError(SymupyError):
    """
        Vehicle Drive Error exception handler, created for handling situations at runtime when the vehicles cannot be created within the network
    
        :return: Vehicle Creation Error 
        :rtype: SymupyDriveVehicleError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)

    def __str__(self) -> str:
        return f"{self.get_messsage}, at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyDriveVehicleError({self.get_messsage},{self._target_dir})"
